/* 
 * copychan.c --
 *
 *	This file contains package Copychan that adds "copychan" command.
 */

#include <tcl.h>
#include <stdlib.h>

#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN

/*
 * VC++ has an alternate entry point called DllMain, so we need to rename
 * our entry point.
 */

#   if defined(_MSC_VER)
#	define EXPORT(a,b) __declspec(dllexport) a b
#	define DllEntryPoint DllMain
#   elif defined(__BORLANDC__)
#	define EXPORT(a,b) a _export b
#   endif
#else
#   define EXPORT(a,b) a b
#endif

/* should use INT_MAX */
#define COPYMAX 0x7fffffff


/*
 * Prototypes for procedures defined later in this file:
 */

static int	CopychanCmd _ANSI_ARGS_((ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv));

EXTERN EXPORT(int,Copychan_Init) _ANSI_ARGS_((Tcl_Interp *interp));



#ifdef __WIN32__
/*
 *----------------------------------------------------------------------
 *
 * DllEntryPoint --
 *
 *	This wrapper function is used by Windows to invoke the
 *	initialization code for the DLL.  If we are compiling
 *	with Visual C++, this routine will be renamed to DllMain.
 *	routine.
 *
 * Results:
 *	Returns TRUE;
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

BOOL APIENTRY
DllEntryPoint(hInst, reason, reserved)
    HINSTANCE hInst;		/* Library instance handle. */
    DWORD reason;		/* Reason this function is being called. */
    LPVOID reserved;		/* Not used. */
{
    return TRUE;
}
#endif

/*
 *----------------------------------------------------------------------
 *
 * TclUnsupported0Cmd --
 *
 *	Implements the "copychannel" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	May copy a chunk from one channel to another.
 *
 *----------------------------------------------------------------------
 */

int
CopychanCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Interpreter in which both channels
                                         * are defined. */
    int argc;				/* How many arguments? */
    char **argv;			/* The argument strings. */
{
    Tcl_Channel inChan, *outChans;
    int requested;
    char *bufPtr, *varName = NULL;
    char buf[128];			/* Buffer to store string
					 * representation of how long
					 * a line was read. */
    int a, next, actuallyRead, actuallyWritten, totalRead, toReadNow, mode;
    int result = TCL_OK;
    int bufSize;			/* Channel buffer size */
    
    /*
     * Assume we want to copy the entire channel.
     */
    
    requested = COPYMAX;

    next = 1;
    if ((next < argc) && (strcmp(argv[next], "-copied") == 0)) {
	next++;
	varName = argv[next];	/* safe even if arg not given */
	next++;
    }

    /*
     * Check if last argument is number of bytes to copy
     */
    if (argc-next > 2 && isdigit((unsigned char) (argv[argc-1][0]))) {
	if (Tcl_GetInt(interp, argv[argc-1], &requested) == TCL_OK) {
	    if (requested < 0) {
		requested = COPYMAX;
	    }
	    argc--;
	} else {
	    return TCL_ERROR;
	}
    }

    if (argc-next < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" ?-copied varName? inChanId outChanId ?outChanId...? ?numBytes?\"",
		(char *) NULL);
        return TCL_ERROR;
    }

    inChan = Tcl_GetChannel(interp, argv[next], &mode);
    if (inChan == (Tcl_Channel) NULL) {
        return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[next],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    next++;	/* consume inChans */
    
    outChans = (Tcl_Channel *) ckalloc((unsigned) argc * sizeof(Tcl_Channel));
    for (a = next; a < argc; a++) {
    	outChans[a-next] = Tcl_GetChannel(interp, argv[a], &mode);
	if (outChans[a-next] == (Tcl_Channel) NULL) {
	    ckfree(outChans);
	    return TCL_ERROR;
	}
	if ((mode & TCL_WRITABLE) == 0) {
	    Tcl_AppendResult(interp, "channel \"", argv[a],
		    "\" wasn't opened for writing", (char *) NULL);
	    ckfree(outChans);
	    return TCL_ERROR;
	}
    }
    
    bufSize = Tcl_GetChannelBufferSize(inChan);
    bufPtr = ckalloc((unsigned) bufSize);
    for (totalRead = 0;
            requested > 0 && result == TCL_OK;
            totalRead += actuallyRead, requested -= actuallyRead) {
        toReadNow = requested;
        if (toReadNow > bufSize) {
            toReadNow = bufSize;
        }
        actuallyRead = Tcl_Read(inChan, bufPtr, toReadNow);
        if (actuallyRead < 0) {
	    Tcl_AppendResult(interp, argv[0], ": ", Tcl_GetChannelName(inChan),
		Tcl_PosixError(interp), (char *) NULL);
	    result = TCL_ERROR;
	    break;
        }
        if (actuallyRead == 0) {
	    break;
        }

	for (a = 0; a < argc-next; a++) {
	    actuallyWritten = Tcl_Write(outChans[a], bufPtr, actuallyRead);
	    if (actuallyWritten < 0) {
		Tcl_AppendResult(interp, argv[0], ": \"",
		    Tcl_GetChannelName(outChans[a]), "\" ",
		    Tcl_PosixError(interp), (char *) NULL);
		result = TCL_ERROR;
		/* don't break, continue looping */
	    }
	    /*
	     * Should do something if (actuallyWritten != actuallyRead)
	     */
	}
    }

    ckfree(bufPtr);
    ckfree(outChans);

    sprintf(buf, "%d", totalRead);
    if (varName) {
        if (Tcl_SetVar(interp, varName, buf, TCL_LEAVE_ERR_MSG) ==
                NULL) {
            return TCL_ERROR;
        }
    }
    if (result == TCL_OK) {
        Tcl_AppendResult(interp, buf, (char *) NULL);
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Copychan_Init --
 *
 *	This is a package initialization procedure, which is called
 *	by Tcl when this package is to be added to an interpreter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

EXPORT(int,Copychan_Init)(interp)
    Tcl_Interp *interp;		/* Interpreter in which the package is
				 * to be made available. */
{
    int code;

    code = Tcl_PkgProvide(interp, "Copychan", "1.0");
    if (code != TCL_OK) {
	return code;
    }
    Tcl_CreateCommand(interp, "copychan", CopychanCmd, (ClientData) 0,
	    (Tcl_CmdDeleteProc *) NULL);
    return TCL_OK;
}

/*eof*/
